import codecs
import copy
import glob
import os
import re
import shutil
import subprocess
import sys
import tempfile


def require(package):
    print('Missing prerequisites. Please run: '
          '"sudo apt install %s"' % package)
    sys.exit(1)

try:
    import polib
except ImportError:
    require('python3-polib')

try:
    import magic
except:
    require('python3-magic')

try:
    import markdown
except:
    require('python3-markdown')

from pelicanconf import PATH, MD_EXTENSIONS, META_TAGS, HIDE_FROM_POT

# This defines how complete we expect translations to be before we
# generate HTML from them. Needs to be string.
TRANSLATION_COMPLETION_PERCENTAGE = '0'

BCP47_OVERRIDES = (
    ('zh_CN', 'zh-hans'),
    ('zh_TW', 'zh-hant'),
)

MD_MAGIC_FILE_TYPES = [
    'ASCII text',
    'UTF-8 Unicode text',
    'UTF-8 Unicode text, with very long lines',
]


def _temp_write_markdown(fn):
    (ret, tmp) = tempfile.mkstemp()
    length = 0
    ret = True
    try:
        markdown.markdownFromFile(fn, extensions=MD_EXTENSIONS, output=tmp)
    except:
        print('Could not convert "%s" into Markdown.' % fn)
        ret = False
    if ret:
        length = len(codecs.open(tmp, 'r', 'utf-8').read())
    os.remove(tmp)
    return (ret, length)


def verify_markdown_file(fn):
    ms = magic.open(magic.MAGIC_NONE)
    ms.load()
    if ms.file(fn) not in MD_MAGIC_FILE_TYPES:
        return False
    (ret, length) = _temp_write_markdown(fn)
    if not length:
        return False
    return ret


def find_bcp47_code(gettext_code):
    if gettext_code not in [c[0] for c in BCP47_OVERRIDES]:
        return gettext_code.lower().replace('_', '-')
    return [c[1] for c in BCP47_OVERRIDES
            if c[0] == gettext_code][0]


class PO4A(object):
    def __init__(self):
        self.default_args = [
            '-f', 'text',
            '-o', 'markdown',
            '-M', 'utf-8',
            ]
        if not shutil.which('po4a'):
            require('po4a')

    def run(self, po4a_command, additional_args, with_output=False):
        args = copy.copy(self.default_args)
        args += additional_args
        if with_output:
            ret = subprocess.Popen([po4a_command]+args, stdout=subprocess.PIPE)
        else:
            ret = subprocess.call([po4a_command]+args)
        return ret

    def gettextize(self, document_fns, pot_file):
        args = copy.copy(self.default_args)
        for document_fn in document_fns:
            args += ['-m', document_fn]
        args += [
            '-p', pot_file,
            '-L', 'utf-8',
            ]
        ret = self.run('po4a-gettextize', args)
        return (not ret)

    def translate(self, doc, po_fn):
        args = [
            '-k', TRANSLATION_COMPLETION_PERCENTAGE,
            '-m', doc,
            '-p', po_fn,
            '-L', 'utf-8',
            ]
        return self.run('po4a-translate', args, with_output=True)


class POFile(object):
    def __init__(self, po_fn):
        self.po_fn = po_fn
        self.pofile = polib.pofile(po_fn)

    def reread(self):
        self.pofile = polib.pofile(self.po_fn)

    def merge(self, pot_file_ob):
        self.pofile.merge(pot_file_ob)

    def save(self):
        self.pofile.save(self.po_fn)

    def find_in_msgid(self, find_str, translated=True, fuzzy=True,
                      untranslated=True):
        entries = []
        if translated:
            entries += self.pofile.translated_entries()
        if fuzzy:
            entries += self.pofile.fuzzy_entries()
        if untranslated:
            entries += self.pofile.untranslated_entries()
        results = []
        for entry in entries:
            if find_str in entry.msgid:
                results += [entry]
        return results

    def hide_attr_list_statements(self):
        entries = []
        for statement in HIDE_FROM_POT:
            entries.extend(self.find_in_msgid(statement))
        statements = r'|'.join(HIDE_FROM_POT)
        for entry in entries:
            matches = re.findall(r'(.*?)\s*?(%s)\s*?' % statements,
                                 entry.msgid)
            # [('How do I update my system?', '!!T')]
            if len(matches) == 1 and len(matches[0]) == 2:
                entry.msgid = matches[0][0]
                entry.comment = matches[0][1]
            if matches[0][1] in entry.msgstr:
                entry.msgstr = entry.msgstr.replace(' %s' % matches[0][1], '')
        self.save()

    def readd_attr_list_statements(self):
        entries = []
        for entry_group in [self.pofile.translated_entries(),
                            self.pofile.fuzzy_entries(),
                            self.pofile.untranslated_entries()]:
            for entry in entry_group:
                for statement in HIDE_FROM_POT:
                    if statement in entry.comment:
                        entries += [entry]
        for entry in entries:
            if not entry.msgid.endswith(entry.comment):
                entry.msgid += ' %s' % entry.comment
            if entry.msgstr and not entry.msgstr.endswith(entry.comment):
                entry.msgstr += ' %s' % entry.comment
            entry.comment = ''
        self.save()

    def safeguard_meta_tags(self):
        for tag in META_TAGS:
            for entry in self.find_in_msgid(tag):
                if entry.msgid == tag:
                    entry.msgstr = entry.msgid
        self.save()

    def find_title_lines(self):
        results = []
        for entry in self.find_in_msgid('Title: '):
            if entry.msgid.startswith('Title: '):
                where = entry.occurrences[0][0]
                first_line = codecs.open(where,
                                         encoding='utf-8').readline().strip()
                results += [(entry, first_line)]
        return results

    def replace_title_lines(self):
        for entry, first_line in self.find_title_lines():
            if entry.msgid != first_line:
                print('Title line "%s" found, but not on the first line '
                      'of "%s".' % (entry.msgid, entry.linenum))
                return False
            entry.msgid = entry.msgid.replace('Title: ', '')
            if self.po_fn.endswith('.po'):
                entry.msgstr = ''
        self.save()
        return True

    def find_link_in_markdown_message(self, entry):
        link_regex = r'\[.+?\]\(\{filename\}(.+?)\).*?'
        link_msgid = re.findall(link_regex, entry.msgid)[0]
        link_msgstr = list(re.findall(link_regex, entry.msgstr))
        return (link_msgid, link_msgstr)

    def rewrite_links(self, documents, bcp47):
        for entry in self.find_in_msgid('{filename}'):
            (link_msgid, link_msgstr) = \
                self.find_link_in_markdown_message(entry)
            if [doc for doc in documents.docs if doc.endswith(link_msgid)]:
                translated_doc_fn = os.path.basename(
                    documents.translated_doc_fn(link_msgid, bcp47))
                if not link_msgstr:
                    entry.msgstr = entry.msgid
                    link_msgstr = [link_msgid]
                entry.msgstr = entry.msgstr.replace(link_msgstr[0],
                                                    translated_doc_fn)
        self.save()

    def find_translated_title_line(self, original_title):
        for entry in self.find_in_msgid(original_title):
            if entry.msgid == original_title:
                if entry.msgstr:
                    return entry.msgstr
                return entry.msgid


class PO(object):
    def __init__(self, po4a):
        self.translations_dir = os.path.abspath(os.path.join(PATH, '../po'))
        self.fake_lang_code = 'en_US'
        self.fake_po_fn = os.path.join(self.translations_dir,
                                       '%s.po' % self.fake_lang_code)
        self.pot_fn = os.path.join(self.translations_dir, 'help.pot')
        self.pot_file_ob = POFile(self.pot_fn)
        self.po4a = po4a
        self.langs = {}
        for po_fn in glob.glob(self.translations_dir+'/*.po'):
            self.add_language(po_fn)

    def add_language(self, po_fn):
        gettext_code = os.path.basename(po_fn).split('.po')[0]
        self.langs[po_fn] = {
            'bcp47': find_bcp47_code(gettext_code),
            'gettext_code': gettext_code,
            'pofile': None,
        }

    def _remove_fake_po_file(self):
        if os.path.exists(self.fake_po_fn):
            os.remove(self.fake_po_fn)

    def __del__(self):
        self._remove_fake_po_file()

    def load_pofile(self, po_fn):
        if not self.langs[po_fn]['pofile']:
            self.langs[po_fn]['pofile'] = POFile(po_fn)

    def gettextize(self, documents):
        if not self.po4a.gettextize(documents.docs, self.pot_fn):
            return False
        self.pot_file_ob.reread()
        return True

    def generate_pot_file(self, documents):
        if not self.gettextize(documents):
            return False
        if not self.pot_file_ob.replace_title_lines():
            return False
        self.pot_file_ob.hide_attr_list_statements()
        for po_fn in self.langs:
            self.load_pofile(po_fn)
            self.langs[po_fn]['pofile'].merge(self.pot_file_ob.pofile)
            if not self.langs[po_fn]['pofile'].replace_title_lines():
                return False
            self.langs[po_fn]['pofile'].hide_attr_list_statements()
        return True

    # we generate a fake translation for en-US which is going to be
    # the default
    def generate_fake_pofile(self):
        self._remove_fake_po_file()
        shutil.copy(self.pot_fn, self.fake_po_fn)
        self.add_language(self.fake_po_fn)

    def find_translated_title_line(self, original_title, po_fn):
        return self.langs[po_fn]['pofile'].find_translated_title_line(
            original_title)

    def rewrite_links(self, documents):
        for po_fn in self.langs:
            self.load_pofile(po_fn)
            self.langs[po_fn]['pofile'].rewrite_links(
                documents, self.langs[po_fn]['bcp47'])

    def safeguard_meta_tags(self):
        for po_fn in self.langs:
            self.load_pofile(po_fn)
            self.langs[po_fn]['pofile'].safeguard_meta_tags()


class Documents(object):
    def __init__(self):
        self.docs = [fn for fn in self.find_docs()
                     if verify_markdown_file(fn)]

    def find_docs(self):
        docs = []
        for dirpath, dirnames, fns in os.walk(PATH):
            docs += [os.path.relpath(os.path.join(dirpath, fn),
                                     os.path.join(PATH, '..'))
                     for fn in fns
                     if fn.endswith('.md')]
        return docs

    def translated_doc_fn(self, fn, bcp47_code):
        match = [doc for doc in self.docs
                 if os.path.basename(doc) == os.path.basename(fn)]
        if not match:
            return None
        return '%s.%s.md' % (match[0].split('.md')[0],
                             bcp47_code)

    def _call_po4a_translate(self, doc, po_fn, po4a):
        res = po4a.translate(doc, po_fn)
        return codecs.decode(res.communicate()[0])

    def write_translated_markdown(self, po, po4a):
        for po_fn in po.langs:
            po.langs[po_fn]['pofile'].readd_attr_list_statements()
            for doc_fn in self.docs:
                output = self._call_po4a_translate(doc_fn, po_fn, po4a)
                title_line = output.split('\n')[0].split('Title: ')[1]
                translated_title_line = po.find_translated_title_line(
                    title_line, po_fn)
                output = '\n'.join([line for line in output.split('\n')][1:])
                new_path = self.translated_doc_fn(doc_fn,
                                                  po.langs[po_fn]['bcp47'])
                text = "Title: %s\nDate:\n\n" % (translated_title_line)
                text += output
                if os.path.exists(new_path):
                    os.remove(new_path)
                if not os.path.exists(os.path.dirname(new_path)):
                    os.makedirs(os.path.dirname(new_path))
                with open(new_path, 'w', encoding='utf-8') as f:
                    f.write(text)
            po.langs[po_fn]['pofile'].hide_attr_list_statements()


class Translations(object):
    def __init__(self):
        self._cleanup()
        self.documents = Documents()
        self.po4a = PO4A()
        self.po = PO(self.po4a)

    def _cleanup(self):
        r = subprocess.Popen(['bzr', 'ignored'], stdout=subprocess.PIPE)
        fns = [os.path.join(PATH, '../..', f.split(' ')[0])
               for f in codecs.decode(r.communicate()[0]).split('\n')
               if f.strip() != '']
        fns = [f for f in fns if os.path.exists(f)]
        for f in fns:
            try:
                shutil.rmtree(f)
            except NotADirectoryError:
                os.remove(f)

    def generate_pot_file(self):
        return self.po.generate_pot_file(self.documents)

    def generate_translations(self):
        self.po.generate_fake_pofile()
        self.po.rewrite_links(self.documents)
        self.po.safeguard_meta_tags()
        self.documents.write_translated_markdown(self.po, self.po4a)
